from sqlalchemy import *

from dataProviderBase import *
from T0Mon.utils import *
from T0Mon.consts import *
from T0Mon.sqlManager import *

from T0Mon.globals import logger
import T0Mon.config

### Module for classes providing data about T0 files ###

class BaseRepackDP(PagedTableDataProvider):

    cols = Cols([
        Col("Id", decorator = buildRepackLink),
        Col("Filesize", decorator = humanFilesize),
        Col("Events"),
        Col("Status"),
        Col("Export status"),
        Col("Dataset"),
        Col("Lfn", decorator = lambda x: textToImageAlt(x))
    ])

    def __init__(self, orderby = 0, asc = True, page = 0):
        PagedTableDataProvider.__init__(self, orderby, asc, page)
        self.title = "Repacked files"

    def _buildCountQuery(self):
        repackedTable = Tbl(TABLES.REPACKED)
        parentageTable = Tbl(TABLES.REPACKED_MERGE_PARENTAGE)

        self.countQuery = \
            select(
                [func.count(repackedTable.c.repacked_id)],
                self.whereCond
            )

    def _buildMainQuery(self):
        repackedTable = Tbl(TABLES.REPACKED)
        rStatusTable = Tbl(TABLES.REPACKED_STATUS)
        eStatusTable = Tbl(TABLES.EXPORT_STATUS)
        datasetTable = Tbl(TABLES.PRIMARY_DATASET)
        parentageTable = Tbl(TABLES.REPACKED_MERGE_PARENTAGE)

        c = (
            repackedTable.c.repacked_id,
            repackedTable.c.filesize,
            repackedTable.c.events,
            rStatusTable.c.status,
            eStatusTable.c.status,
            datasetTable.c.name,
            repackedTable.c.lfn
        )

        if self.asc: o = c[self.orderby].asc()
        else: o = c[self.orderby].desc()

        self.mainQuery = select(
            c,
            self.whereCond,
            from_obj = [repackedTable.
                            outerjoin(rStatusTable, rStatusTable.c.id == repackedTable.c.status).
                            outerjoin(eStatusTable, repackedTable.c.export_status == eStatusTable.c.id).
                            outerjoin(datasetTable, datasetTable.c.dataset_id == repackedTable.c.dataset_id)
                        ]
            ).order_by(o)

class RepackByIdsDP(BaseRepackDP):
    def __init__(self, ids, orderby = 0, asc = True, page = 0):
        BaseRepackDP.__init__(self, orderby, asc, page)

        repackedTable = Tbl(TABLES.REPACKED)

        self.whereCond = repackedTable.c.repacked_id.in_(ids)

        BaseRepackDP._load(self)

class RepackByRunIdDP(BaseRepackDP):
    def __init__(self, runid, orderby = 0, asc = True, page = 0):
        BaseRepackDP.__init__(self, orderby, asc, page)

        repackedTable = Tbl(TABLES.REPACKED)
        parentageTable = Tbl(TABLES.REPACKED_MERGE_PARENTAGE)

        self.whereCond = and_(
            repackedTable.c.run_id == runid,
            not_(repackedTable.c.repacked_id.in_(select([parentageTable.c.output_id])))
        )
        BaseRepackDP._load(self)

class RepackByRepackMergeIdsDP(BaseRepackDP):
    def __init__(self, ids, orderby = 0, asc = True, page = 0):
        BaseRepackDP.__init__(self, orderby, asc, page)

        repackedTable = Tbl(TABLES.REPACKED)
        parentageTable = Tbl(TABLES.REPACKED_MERGE_PARENTAGE)

        self.whereCond = \
            or_(
                repackedTable.c.repacked_id.in_(select([parentageTable.c.input_id], parentageTable.c.output_id.in_(ids))),
                and_(
                    repackedTable.c.repacked_id.in_(ids),
                    not_(repackedTable.c.repacked_id.in_(select([parentageTable.c.output_id])))
                )
            )
        BaseRepackDP._load(self)

class RepackByRepackJobIdsDP(BaseRepackDP):
    def __init__(self, ids, orderby = 0, asc = True, page = 0):
        BaseRepackDP.__init__(self, orderby, asc, page)

        repackedTable = Tbl(TABLES.REPACKED)
        assocTable = Tbl(TABLES.REPACK_STREAMER_ASSOC)
        jobAssocTable = Tbl(TABLES.JOB_DATASET_STREAMER_ASSOC)
        parentageTable = Tbl(TABLES.REPACKED_MERGE_PARENTAGE)

        self.whereCond = and_(
            repackedTable.c.repacked_id.in_(
                select ([assocTable.c.repacked_id],
                    assocTable.c.streamer_id.in_(
                        select([jobAssocTable.c.streamer_id],
                            jobAssocTable.c.job_id.in_(ids))
                    )
                )
            ),
            repackedTable.c.dataset_id.in_(select([jobAssocTable.c.dataset_id], jobAssocTable.c.job_id.in_(ids))),
            not_(repackedTable.c.repacked_id.in_(select([parentageTable.c.output_id])))
        )
        BaseRepackDP._load(self)

class BaseRepackMergeDP(PagedTableDataProvider):

    cols = Cols([
        Col("Id", decorator = buildRepackMergeLink),
        Col("Filesize", decorator = humanFilesize),
        Col("Events"),
        Col("Status"),
        Col("Export status"),
        Col("Dataset"),
        Col("Lfn", decorator = lambda x: textToImageAlt(x))
    ])

    def __init__(self, orderby = 0, asc = True, page = 0):
        PagedTableDataProvider.__init__(self, orderby, asc, page)
        self.title = "Merge files"

    def _buildCountQuery(self):
        repackedTable = Tbl(TABLES.REPACKED)

        self.countQuery = select([func.count('*')],
            self.whereCond
        )

    def _buildMainQuery(self):
        repackedTable = Tbl(TABLES.REPACKED)
        rStatusTable = Tbl(TABLES.REPACKED_STATUS)
        eStatusTable = Tbl(TABLES.EXPORT_STATUS)
        datasetTable = Tbl(TABLES.PRIMARY_DATASET)

        c = (
            repackedTable.c.repacked_id,
            repackedTable.c.filesize,
            repackedTable.c.events,
            rStatusTable.c.status,
            eStatusTable.c.status,
            datasetTable.c.name,
            repackedTable.c.lfn
        )

        if self.asc: o = c[self.orderby].asc()
        else: o = c[self.orderby].desc()

        self.mainQuery = select(
            c,
            self.whereCond,
            from_obj = [repackedTable.
                            outerjoin(rStatusTable, rStatusTable.c.id == repackedTable.c.status).
                            outerjoin(eStatusTable, repackedTable.c.export_status == eStatusTable.c.id).
                            outerjoin(datasetTable, datasetTable.c.dataset_id == repackedTable.c.dataset_id)
                        ]
            ).order_by(o)

class RepackMergeByRecoJobIdsDP(BaseRepackMergeDP):
    def __init__(self, ids, orderby = 0, asc = True, page = 0):
        BaseRepackMergeDP.__init__(self, orderby, asc, page)
        repackedTable = Tbl(TABLES.REPACKED)
        assocTable = Tbl(TABLES.RECO_JOB_REPACK_ASSOC)
        self.whereCond = repackedTable.c.repacked_id.in_(
            select([assocTable.c.repacked_id], assocTable.c.job_id.in_(ids))
        )
        BaseRepackMergeDP._load(self)

class RepackMergeByIdsDP(BaseRepackMergeDP):
    def __init__(self, ids, orderby = 0, asc = True, page = 0):
        BaseRepackMergeDP.__init__(self, orderby, asc, page)
        repackedTable = Tbl(TABLES.REPACKED)
        self.whereCond = repackedTable.c.repacked_id.in_(ids)
        BaseRepackMergeDP._load(self)

class RepackMergeByRepackMergeJobIdsDP(BaseRepackMergeDP):
    def __init__(self, ids, orderby = 0, asc = True, page = 0):
        BaseRepackMergeDP.__init__(self, orderby, asc, page)
        repackedTable = Tbl(TABLES.REPACKED)
        assocTable = Tbl(TABLES.MERGE_JOB_REPACK_ASSOC)
        parentageTable = Tbl(TABLES.REPACKED_MERGE_PARENTAGE)
        self.whereCond = repackedTable.c.repacked_id.in_(
            select([parentageTable.c.output_id], parentageTable.c.input_id.in_(
                select([assocTable.c.repacked_id], assocTable.c.job_id.in_(ids))
            ))
        )
        BaseRepackMergeDP._load(self)

class RepackMergeByRepackIdsDP(BaseRepackMergeDP):
    def __init__(self, ids, orderby = 0, asc = True, page = 0):
        BaseRepackMergeDP.__init__(self, orderby, asc, page)
        repackedTable = Tbl(TABLES.REPACKED)
        assocTable = Tbl(TABLES.MERGE_JOB_REPACK_ASSOC)
        parentageTable = Tbl(TABLES.REPACKED_MERGE_PARENTAGE)
        self.whereCond = \
            and_(
                or_(
                    repackedTable.c.status == REPACK_RECONSTRUCTION_QUEUED,
                    repackedTable.c.status == REPACK_RECONSTRUCTION_DISABLED
                ),
                or_(
                    repackedTable.c.repacked_id.in_(
                        select([parentageTable.c.output_id],
                            parentageTable.c.input_id.in_(ids)
                        )
                    ),
                    repackedTable.c.repacked_id.in_(ids)
                )
            )
        if len(ids) != 0:
            BaseRepackMergeDP._load(self)

class RepackMergeByRunIdDP(BaseRepackMergeDP):
    def __init__(self, runid, orderby = 0, asc = True, page = 0):
        BaseRepackMergeDP.__init__(self, orderby, asc, page)
        repackedTable = Tbl(TABLES.REPACKED)
        self.whereCond = \
            and_(
                or_(
                    repackedTable.c.status == REPACK_RECONSTRUCTION_QUEUED,
                    repackedTable.c.status == REPACK_RECONSTRUCTION_DISABLED
                ),
                repackedTable.c.run_id == runid
            )
        BaseRepackMergeDP._load(self)

class BaseRecoDP(PagedTableDataProvider):

    cols = Cols([
        Col("Id", decorator=buildRecoLink),
        Col("Filesize", decorator=humanFilesize),
        Col("Events"),
        Col("Status"),
        Col("Export status"),
        Col("Dataset"),
        Col("Lfn", decorator = lambda x: textToImageAlt(x))
    ])

    def __init__(self, orderby = 0, asc = True, page = 0):
        PagedTableDataProvider.__init__(self, orderby, asc, page)
        self.title = "Reco files"

    def _buildCountQuery(self):
        recoTable = Tbl(TABLES.RECONSTRUCTED)

        self.countQuery = select(
            [func.count(recoTable.c.reconstructed_id)],
            self.whereCond
        )

    def _buildMainQuery(self):
        recoTable = Tbl(TABLES.RECONSTRUCTED)
        rStatusTable = Tbl(TABLES.RECONSTRUCTED_STATUS)
        eStatusTable = Tbl(TABLES.EXPORT_STATUS)
        datasetTable = Tbl(TABLES.PRIMARY_DATASET)

        c = (
            recoTable.c.reconstructed_id,
            recoTable.c.filesize,
            recoTable.c.events,
            rStatusTable.c.status,
            eStatusTable.c.status,
            datasetTable.c.name,
            recoTable.c.lfn
        )

        if self.asc: o = c[self.orderby].asc()
        else: o = c[self.orderby].desc()

        self.mainQuery = select(
            c,
            self.whereCond,
            from_obj = [
                recoTable.
                outerjoin(rStatusTable, rStatusTable.c.id == recoTable.c.status).
                outerjoin(eStatusTable, recoTable.c.export_status == eStatusTable.c.id).
                outerjoin(datasetTable, datasetTable.c.dataset_id == recoTable.c.dataset_id)
            ]
            ).order_by(o)

class RecoByIdsDP(BaseRecoDP):
    def __init__(self, ids, orderby = 0, asc = True, page = 0):
        BaseRecoDP.__init__(self, orderby, asc, page)
        recoTable = Tbl(TABLES.RECONSTRUCTED)
        self.whereCond = recoTable.c.reconstructed_id.in_(ids)
        BaseRecoDP._load(self)

class RecoByRunIdDP(BaseRecoDP):
    def __init__(self, runid, orderby = 0, asc = True, page = 0):
        BaseRecoDP.__init__(self, orderby, asc, page)
        recoTable = Tbl(TABLES.RECONSTRUCTED)
        self.whereCond = recoTable.c.run_id == runid
        BaseRecoDP._load(self)

class RecoByRecoJobIdsDP(BaseRecoDP):
    def __init__(self, ids, orderby = 0, asc = True, page = 0):
        BaseRecoDP.__init__(self, orderby, asc, page)
        recoTable = Tbl(TABLES.RECONSTRUCTED)
        assocTable = Tbl(TABLES.RECO_JOB_REPACK_ASSOC)
        parentageTable = Tbl(TABLES.REPACKED_RECO_PARENTAGE)
        self.whereCond = recoTable.c.reconstructed_id.in_(
            select([parentageTable.c.output_id], parentageTable.c.input_id.in_(
                select([assocTable.c.repacked_id], assocTable.c.job_id.in_(ids))
            ))
        )
        BaseRecoDP._load(self)

class StreamerDP(PagedTableDataProvider):

    cols = Cols([
        Col("Id", decorator = buildStreamerLink),
        Col("Lumi Id"),
        Col("Filesize", decorator = humanFilesize),
        Col("Events"),
        Col("Name"),
        Col("E"),
        Col("S"),
        Col("Lfn", decorator = lambda x: textToImageAlt(x))
    ])

    def __init__(self, orderby = 0, asc = True, page = 0):
        PagedTableDataProvider.__init__(self, orderby, asc, page)
        self.title = 'Streamer files'

    def _buildCountQuery(self):
        streamerTable = Tbl(TABLES.STREAMER)

        self.countQuery = select(
            [func.count(streamerTable.c.streamer_id)],
            self.whereCond
        )

    def _buildMainQuery(self):
        streamerTable = Tbl(TABLES.STREAMER)

        c = (
            streamerTable.c.streamer_id,
            streamerTable.c.lumi_id,
            streamerTable.c.filesize,
            streamerTable.c.events,
            streamerTable.c.streamname,
            streamerTable.c.exportable,
            streamerTable.c.splitable,
            streamerTable.c.lfn
        )

        if self.asc: o = c[self.orderby].asc()
        else: o = c[self.orderby].desc()

        self.mainQuery = select(
            c,
            self.whereCond
        ).order_by(o)

class StreamerByRunIdDP(StreamerDP):
    def __init__(self, runid, orderby = 0, asc = True, page = 0):
        StreamerDP.__init__(self, orderby, asc, page)
        streamerTable = Tbl(TABLES.STREAMER)
        self.whereCond = (streamerTable.c.run_id == runid)
        StreamerDP._load(self)

class StreamerByIdsDP(StreamerDP):
    def __init__(self, ids, orderby = 0, asc = True, page = 0):
        StreamerDP.__init__(self, orderby, asc, page)
        streamerTable = Tbl(TABLES.STREAMER)
        self.whereCond = (streamerTable.c.streamer_id.in_(ids))
        StreamerDP._load(self)

class StreamerByRepackJobIdsDP(StreamerDP):
    def __init__(self, ids, orderby = 0, asc = True, page = 0):
        StreamerDP.__init__(self, orderby, asc, page)
        streamerTable = Tbl(TABLES.STREAMER)
        assocTable = Tbl(TABLES.JOB_DATASET_STREAMER_ASSOC)
        self.whereCond = streamerTable.c.streamer_id.in_(
            select([func.distinct(assocTable.c.streamer_id)], assocTable.c.job_id.in_(ids))
        )
        StreamerDP._load(self)
